@@ -2,7 +2,7 @@ from typing import NamedTuple
 
 from pygments.token import _TokenType
 
-HTTP_METHODS = ['get', 'post', 'put', 'delete']
+HTTP_METHODS = ['get', 'post', 'put', 'delete', 'head']
 AUTO_SAVE_NAME = '__auto__'
 DEFAULT_SAVE_NAME = '__default__'
 
