@@ -1,17 +1,32 @@
+import base64
+import functools
 import json
 import logging
 import os
 from abc import ABCMeta, abstractmethod
-from typing import List
+from typing import List, Iterable
 
+import elasticsearch.client.utils
 from configobj import Section
 from elasticsearch import Elasticsearch, AuthenticationException
 from urllib3 import Timeout
 
+from peek.config import get_global_config
 from peek.errors import PeekError
 
 _logger = logging.getLogger(__name__)
 
+# Wrappers for adding request and response content type
+_wrapper_json = elasticsearch.client.utils.query_params(
+    request_mimetypes=["application/json"],
+    response_mimetypes=["application/json"]
+)
+
+_wrapper_text = elasticsearch.client.utils.query_params(
+    request_mimetypes=["application/json"],
+    response_mimetypes=["text/plain", "application/json"]
+)
+
 
 class NoopDeserializer:
     def __init__(self):
@@ -84,6 +99,8 @@ class EsClient(BaseClient):
             headers=headers,
             ssl_assert_hostname=assert_hostname,
         )
+        # skip product verification because it requires deserialization of payload
+        self.es.transport._verified_elasticsearch = True
 
     def perform_request(self, method, path, payload=None, deserialize_it=False, **kwargs):
         _logger.debug(f'Performing request: {method!r}, {path!r}, {payload!r}')
@@ -92,11 +109,23 @@ class EsClient(BaseClient):
             if not deserialize_it:
                 # Avoid deserializing the response since we parse it with the main loop for syntax highlighting
                 self.es.transport.deserializer = noopDeserializer
-            return self.es.transport.perform_request(method, path, body=payload, **kwargs)
+            _wrapper = self._get_wrapper(path)
+            return _wrapper(functools.partial(self.es.transport.perform_request, method, path))(body=payload, **kwargs)
         finally:
             if not deserialize_it:
                 self.es.transport.deserializer = deserializer
 
+    def _get_wrapper(self, path):
+        if path == '/_nodes/hot_threads':
+            return _wrapper_text
+        elif path.startswith("/_cat/"):
+            if get_global_config().as_bool('accept_json_for_cat'):
+                return _wrapper_json
+            else:
+                return _wrapper_text
+        else:
+            return _wrapper_json
+
     def info(self):
         if self.api_key:
             auth = f'ApiKey {self.api_key[0][:10]}...'
@@ -252,16 +281,37 @@ class RefreshingEsClient(BaseClient):
         )
 
 
+class DelegatingListener:
+
+    def __init__(self, on_add=None, on_set=None, on_remove=None):
+        self._on_add = on_add
+        self._on_set = on_set
+        self._on_remove = on_remove
+
+    def on_add(self, m):
+        return self._on_add(m) if self._on_add is not None else True
+
+    def on_set(self, m):
+        return self._on_set(m) if self._on_set is not None else True
+
+    def on_remove(self, m, c):
+        return self._on_remove(m, c) if self._on_remove is not None else True
+
+
 class EsClientManager:
 
-    def __init__(self):
+    def __init__(self, listeners: Iterable[DelegatingListener] = ()):
         self._clients: List[EsClient] = []
         self._index_current = None
+        self.listeners = listeners
 
     def add(self, client):
         self._clients.append(client)
         self._index_current = len(self._clients) - 1
         # TODO: maintain size
+        for listener in self.listeners:
+            if listener.on_add(self) is False:
+                break
 
     @property
     def index_current(self):
@@ -281,10 +331,13 @@ class EsClientManager:
             self._index_current = self._clients.index(self.get_client(name))
         elif isinstance(x, int):
             if x < 0 or x >= len(self._clients):
-                raise PeekError(f'Attempt to set ES client at invalid index [{x}]')
+                raise PeekError(f'Attempt to set ES client at invalid index {x!r}')
             self._index_current = x
         else:
             raise ValueError(f'Connection must be specified by either name or index, got {x!r}')
+        for listener in self.listeners:
+            if listener.on_set(self) is False:
+                break
 
     def clients(self):
         return self._clients
@@ -315,11 +368,9 @@ class EsClientManager:
             idx = self._clients.index(self.get_client(x))
             self.remove_client(idx)
         elif isinstance(x, int):
-            if len(self._clients) == 1:
-                raise PeekError('Cannot delete the last connection')
             if x < 0 or x >= len(self._clients):
                 raise PeekError(f'Attempt to remove ES client at invalid index [{x}]')
-            self._clients.pop(x)
+            removed = self._clients.pop(x)
             if not self._clients:
                 self._index_current = None
                 return
@@ -327,6 +378,9 @@ class EsClientManager:
                 self._index_current -= 1
             elif x == self._index_current:
                 self._index_current = 0
+            for listener in self.listeners:
+                if listener.on_remove(self, removed) is False:
+                    break
         else:
             raise ValueError(f'Connection must be specified by either name or index, got {x!r}')
 
@@ -338,6 +392,17 @@ class EsClientManager:
         self._index_current = 0
         self._clients = [keep]
 
+    def move_current_to(self, idx: int):
+        if not isinstance(idx, int):
+            raise PeekError(f'Index must be integer, got {idx!r}')
+        if 0 <= idx < len(self._clients):
+            if idx == self._index_current:
+                return
+            self._clients.insert(idx, self._clients.pop(self._index_current))
+            self._index_current = idx
+        else:
+            raise PeekError(f'Attempt to move ES client to an invalid index: {idx!r}')
+
     def to_dict(self):
         result = {
             '_index_current': self._index_current,
@@ -410,8 +475,7 @@ def connect(app, **options):
     if isinstance(app.config.get('connection'), Section):
         final_options.update({k: v for k, v in app.config.get('connection').dict().items() if v})
 
-    # Override with provided options, including null values since it could be intentional
-    final_options.update({k: v for k, v in options.items()})
+    final_options.update(_maybe_configure_smart_connect(app, options))
 
     if final_options['cloud_id'] is not None:
         final_options['hosts'] = None
@@ -424,6 +488,60 @@ def connect(app, **options):
         return _connect_userpass(app, **final_options)
 
 
+def _maybe_configure_smart_connect(app, options: dict):
+    # Override with provided options, including null values since it could be intentional
+    smart_options = {k: v for k, v in options.items()}
+    # Attempt to configure smart connect based on the last HTTP response
+    if len(options) == 0:
+        try:
+            last_response = app.vm.get_value('_')
+            if 'id' in last_response and 'api_key' in last_response:
+                _maybe_copy_current_client_options(app, smart_options)
+                smart_options['api_key'] = f'{last_response["id"]}:{last_response["api_key"]}'
+            elif 'access_token' in last_response:
+                _maybe_copy_current_client_options(app, smart_options)
+                smart_options['token'] = last_response['access_token']
+            elif 'token' in last_response and 'value' in last_response['token']:
+                _maybe_copy_current_client_options(app, smart_options)
+                smart_options['token'] = last_response['token']['value']
+            else:
+                last_request = app.vm.get_value('__')
+                urlpath = last_request['path']
+                if last_request['method'] in ('POST', 'PUT') and (urlpath.startswith('/_security/user/')
+                                                                  or urlpath.startswith('/_xpack/security/user/')):
+                    if urlpath.startswith('/_security/user/'):
+                        username = urlpath[len('/_security/user/'):]
+                    else:
+                        username = urlpath[len('/_xpack/security/user/'):]
+                    payload = json.loads(last_request['payload'])
+                    password = payload.get('password', None)
+                    _maybe_copy_current_client_options(app, smart_options)
+                    smart_options['username'] = username
+                    smart_options['password'] = password
+
+        except NameError:  # if _ does not exist, simply ignore and proceed
+            pass
+
+    return smart_options
+
+
+def _maybe_copy_current_client_options(app, options: dict):
+    if hasattr(app, 'es_client_manager') and len(app.es_client_manager.clients()) > 0:
+        current_es_client = app.es_client_manager.current
+        if current_es_client.cloud_id is not None:
+            options['cloud_id'] = current_es_client.cloud_id
+        else:
+            options['hosts'] = current_es_client.hosts
+
+        options['use_ssl'] = current_es_client.use_ssl
+        options['verify_certs'] = current_es_client.verify_certs
+        options['assert_hostname'] = current_es_client.assert_hostname
+        options['ca_certs'] = current_es_client.ca_certs
+        options['client_cert'] = current_es_client.client_cert
+        options['client_key'] = current_es_client.client_key
+        # not copy the headers
+
+
 def _connect_userpass(app, **options):
     username = options.get('username')
     password = options.get('password')
@@ -474,11 +592,18 @@ def _connect_userpass(app, **options):
 
 def _connect_api_key(app, **options):
     _logger.debug('Connecting with API key')
+    api_key = options['api_key']
+    if ':' not in api_key:
+        api_key = base64.decodebytes(api_key.encode()).decode()
+
+    if ':' not in api_key:
+        raise ValueError('invalid api key credential format')
+
     return EsClient(
         name=options['name'],
         hosts=options['hosts'],
         cloud_id=options['cloud_id'],
-        api_key=options['api_key'].split(':'),
+        api_key=api_key.split(':'),
         use_ssl=options['use_ssl'],
         verify_certs=options['verify_certs'],
         assert_hostname=options['assert_hostname'],
@@ -530,9 +655,26 @@ def _keyring(service_name, key, value=None):
 
 class ConnectFunc:
     def __call__(self, app, **options):
-        app.es_client_manager.add(connect(app, **options))
+        test_connection = app.config.as_bool('test_connection')
+        opt = options.pop('test', None)
+        if opt is not None:
+            test_connection = opt
+        es_client = connect(app, **options)
+        if test_connection:
+            try:
+                app.display.info(es_client.perform_request('GET', '/_security/_authenticate'))
+            except Exception as e:
+                app.display.error(e)
+                return
+        app.es_client_manager.add(es_client)
         return str(app.es_client_manager)
 
     @property
     def options(self):
-        return dict(DEFAULT_OPTIONS)
+        _options = dict(DEFAULT_OPTIONS)
+        _options['test'] = None
+        return _options
+
+    @property
+    def description(self):
+        return 'Connect to an Elasticsearch cluster'
